defmodule Dp.Kif do

  @moduledoc """
  Kifejezések

  @author "hanak@emt.bme.hu"
  @date   "$LastChangedDate: 2022-09-06 09:48:24 +0200 (k, 06 szept 2022) $$"
  """

  # --------------------- BEMELEGÍTÉS ------------------------------
  
  @type fa :: 1 | {fa, '+', fa}
  @spec fak(n :: integer) :: fs :: [fa]
  # Az összes n levelű fa listája fs
  def fak(1), do: [1]
  def fak(n) do
    # \ a folytatódó sor jelzésére, ha a for után meg akarjuk törni a sort
    # Mivel a \-nek más jelentése is van az Elixirben, utána nem állhat más,
    # pl. szóköz vagy horizontális tabulátor, csak újsor karakter
    for \
        i   <- (Enum.to_list 1..(n-1)),
        bfa <- fak(i),
        jfa <- fak(n-i),
      do: {bfa, '+', jfa}
  end

  # --------------------- KIFEJEZÉSFA ------------------------------

  @type muv :: '+' | '-' | '*' | '/'
  @type kif :: integer | {kif, muv, kif}
  @spec kifek(xs :: [integer]) :: ks :: [kif]
  # Az xs listában lévő számokból alapműveletekkel előálló
  # összes lehetséges kifejezés listája ks, ahol a levelek
  # in-order sorrendje az xs-beli sorrend
  # Tömören:
  # Az xs-beli számokból épített kif-ek listája ks
  # Megjegyzés: az Elixirben explicite fel kell sorolnunk,
  # nincs automatikus visszalépés
  def kifek([x]), do: [x]
  def kifek(xs),  do:
    for {xbs, xjs} <- kettevagasok(xs),
        b          <- kifek(xbs),
        j          <- kifek(xjs),
        m          <- ['+', '-', '*', '/'],
      do: {b, m, j}

  @spec kettevagasok(xs :: [integer]) :: pls :: [{bls :: [integer], jls :: [integer]}]
  # Az összes olyan nemüres bls és jls listákból álló párok listája
  # pls, amelyek páronként összefűzve kiadják az xs listát 
  def kettevagasok(xs), do:
    for i          <- 1..((length xs) - 1),
        {bls, jls} <- [Enum.split(xs, i)],
      do: {bls, jls}

  @spec ertek(k :: kif) :: e :: integer
  # A k kifejezés értéke e
  def ertek({bfa, mjel, jfa}), do: muv(mjel).(ertek(bfa), ertek(jfa))
  def ertek(i), do: i   # A klózok sorrendje itt számít, fontos

  @type optype :: (number, number -> number)
  @spec muv(op :: charlist) :: m::optype
  def muv('+'), do: &+/2  
  def muv('-'), do: &-/2  
  def muv('*'), do: &*/2  
  def muv('/'), do: &//2  

  @spec ertek2(k :: kif) :: e :: integer
  # A k kifejezés értéke e
  def ertek2({bfa, mjel, jfa}) do
    op = fn '+' -> &+/2; '-' -> &-/2; '*' -> &*/2; '/' -> &//2 end
    op.(mjel).(ertek2(bfa), ertek2(jfa))
    end
  def ertek2(i), do: i   # A klózok sorrendje itt számít, fontos!

  @spec permutaciok(xs :: [any]) :: pss :: [[any]]
  # Az xs lista elemeinek összes permutációját tartalmazó lista pss
  def permutaciok([]), do: [[]]
  def permutaciok(xs), do:
    for y <- xs, ys <- permutaciok(xs -- [y]), do: [y|ys]
 
  @spec megoldasok(xs :: [integer], e :: integer) :: ks :: [kif]
  # Az xs számokkal az e eredményt adó kifejezések listája ks
  def megoldasok(xs, ered), do:
    for pss <- permutaciok(xs),
        kif <- kifek(pss),
        (try do ertek(kif) catch _,_ -> nil end) == ered,
      do: kif
  
  @spec megoldasok() :: ks :: [kif]
  # Az [1,3,4,6] számokkal az 24 eredményt adó kifejezések listája ks
  def megoldasok(), do: megoldasok([1, 3, 4, 6], 24)

# --------------------- TESZTEK ------------------------------

# pelda(1) ->
#     [X || X <- [1,2,3]];
# pelda(2) ->
#     [X || X <- [1,2,3], X*X > 5];
# pelda(3) ->
#     lists:seq(1,3);
# pelda(4) ->
#     [{X,Y} || X <- [1,2,3,4], Y <- lists:seq(1,X)];
# pelda(5) ->
#     fak(4);
# pelda(6) ->
#     length(fak(4));
# pelda(7) ->
#     kifek([1,3,4]);
# pelda(8) ->
#     ertek({{1, '-', {3, '*', 4} }, '+', 6});
# pelda(9) ->
#     permutaciok([1,3,4]);
# pelda(10) ->
#     L = [-1,-2,-3,-4],
#     [{Elso,Masodik} || I <- [1,2,3], {Elso,Masodik} <- [lists:split(I, L)]].
  
end
